==> conc-2.0.1.tar.gz: conc-2.0.1/mgr/cgroup.c <==
#include <string.h>
#include "cgroup.h"
#include "console.h"
#include "config-file.h"

struct _CGroup {
    gchar *name;
    Manager *manager;
    ConfigPath config_path;
    GList *consoles;
    Status status;
};

static void
cgroup_tree_free(gpointer data) {
    CGroup cg = data;

    if (cg->name != NULL) {
	g_free(cg->name);
    }
    if (cg->config_path != NULL) {
	config_path_free(cg->config_path);
    }
    while (cg->consoles != NULL) {
	g_free(cg->consoles->data);
	cg->consoles = g_list_remove(cg->consoles, cg->consoles->data);
    }
}

static gchar *
group_member(CGroup cg,
	const gchar *console) {
    GList *l;

    for (l = cg->consoles; l; l = g_list_next(l)) {
	if (!strcmp(l->data, console)) {
	    return l->data;
	}
    }
    return NULL;
}

static void
cgroup_config_write(CGroup cg, gboolean flush) {
    ConfigPath cgroups;
    GList *l;

    g_assert(cg);

    if (strcmp(cg->name, "ALL") == 0) {
	return;
    }
    cgroups = config_path(cg->manager->config, "cgroups", NULL);
    if (cgroups == NULL) {
	ConfigPath root = config_path(cg->manager->config, NULL);

	cgroups = config_path_create(cg->manager->config, root, "cgroups");
	config_path_free(root);
    }
    if (cg->config_path != NULL) {
	config_delete(cg->config_path);
    }
    cg->config_path = config_path_create(cg->manager->config,
	    cgroups, "cgroup");
    config_path_free(cgroups);
    config_set(cg->config_path, "name", cg->name);
    for (l = cg->consoles; l != NULL; l = g_list_next(l)) {
	config_set(cg->config_path, "console", l->data);
    }
    if (flush) {
	config_file_write(cg->manager->config, NULL);
    }
}

static void
add_console(CGroup cg,
	const gchar *console) {
    g_assert(cg);
    g_assert(console);

    if (group_member(cg, console) != NULL) {
	return;
    }
    cg->consoles = g_list_prepend(cg->consoles, g_strdup(console));
    event_raise(cg->manager, EVENT_CGROUP_CHANGED, cg->name, console, 1);
}

static void
remove_console(CGroup cg,
	const gchar *console) {
    gchar *m;
    gchar *c = g_strdup(console);

    g_assert(cg);
    g_assert(console);

    m = group_member(cg, console);
    if (m == NULL) {
	return;
    }
    cg->consoles = g_list_remove(cg->consoles, m);
    g_free(m);
    event_raise(cg->manager, EVENT_CGROUP_CHANGED, cg->name, c, 0);
    g_free(c);
}

static void
console_rename(CGroup cg,
	const gchar *oldname,
	const gchar *newname) {
    gchar *m;

    g_assert(cg);
    g_assert(oldname);
    g_assert(newname);

    m = group_member(cg, oldname);
    if (m) {
	cg->consoles = g_list_remove(cg->consoles, m);
	g_free(m);
	cg->consoles = g_list_prepend(cg->consoles, g_strdup(newname));
	cgroup_config_write(cg, TRUE);
    }
}

static void
console_added(Manager *manager,
	const EventData *event,
	gpointer data) {
    CGroup all;

    all = g_tree_lookup(manager->cgroups, "ALL");
    g_assert(all);
    add_console(all, event->console_added.console);
    cgroup_config_write(all, TRUE);
}

static gboolean
console_removed_fe(gpointer key,
	gpointer value,
	gpointer data) {
    CGroup cg = value;
    gchar *console = data;

    remove_console(cg, console);
    cgroup_config_write(cg, FALSE);
    return FALSE;
}

static void
console_removed(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_tree_foreach(manager->cgroups, console_removed_fe,
	    (gchar *)event->console_removed.console);
}

static gboolean
console_renamed_fe(gpointer key,
	gpointer value,
	gpointer data) {
    CGroup cg = value;
    EventData *event = data;

    console_rename(cg, event->console_renamed.console,
	    event->console_renamed.newname);
    return FALSE;
}

static void
console_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_tree_foreach(manager->cgroups, console_renamed_fe,
	    (gpointer)event);
}

extern void
cgroup_load(Manager *manager) {
    ConfigPath cp;
    GList *cgroups, *l;
    CGroup all;


    g_assert(manager);
    g_assert(manager->config);

    manager->cgroups = g_tree_new_full((GCompareDataFunc)strcmp,
	    NULL, g_free, cgroup_tree_free);

    cp = config_path(manager->config, "cgroups", NULL);
    if (cp == NULL) {
	cgroups = NULL;
    } else {
	cgroups = config_children(cp);
	config_path_free(cp);
    }
    for (l = cgroups; l != NULL; l = g_list_next(l)) {
	const gchar *str;
	GList *consoles;
	CGroup cg;

	cp = l->data;
	str = config_get_value(cp, "name");
	if (str == NULL) {
	    g_printerr("configuration file corrupt "
		    "(anonymous console group)\n");
	    continue;
	}
	if (strcmp(str, "ALL") == 0) {
	    /* ALL is reconstructed later */
	    continue;
	}
	cg = g_new0(struct _CGroup, 1);
	cg->name = g_strdup(str);
	cg->manager = manager;
	cg->config_path = cp;
	cg->consoles = NULL;
	cg->status = STATUS_OK;

	consoles = config_children(cp);
	while (consoles != NULL) {
	    ConfigPath ccp = consoles->data;
	    const ConfigDirective *cd = config_get(ccp);

	    consoles = g_list_remove(consoles, ccp);
	    config_path_free(ccp);
	    if (strcmp(cd->name, "console") != 0 ||
		    cd->value == NULL) {
		continue;
	    }
	    cg->consoles = g_list_prepend(cg->consoles, g_strdup(cd->value));
	}
	g_tree_insert(manager->cgroups, g_strdup(cg->name), cg);
    }

    all = g_new0(struct _CGroup, 1);
    all->name = g_strdup("ALL");
    all->manager = manager;
    all->config_path = NULL;
    all->consoles = console_list(manager);
    g_tree_insert(manager->cgroups, g_strdup("ALL"), all);

    event_register(manager, EVENT_CONSOLE_ADDED, console_added, NULL);
    event_register(manager, EVENT_CONSOLE_REMOVED, console_removed, NULL);
    event_register(manager, EVENT_CONSOLE_RENAMED, console_renamed, NULL);
}

extern GList *
cgroup_list_consoles(CGroup g) {
    GList *p = NULL;
    GList *l;

    for (l = g->consoles; l != NULL; l = g_list_next(l)) {
	p = g_list_prepend(p, g_strdup(l->data));
    }
    return p;
}

static gboolean
check_status_fe(gpointer key,
        gpointer value,
        gpointer data) {
    Manager *manager = data;
    CGroup cg = value;
    GList *l;
    Status s = STATUS_OK;

    for (l = cg->consoles; l != NULL; l = g_list_next(l)) {
	Status cs;
	Console c = g_tree_lookup(manager->consoles, l->data);

	if (c == NULL) {
	    continue;
	}
	cs = console_get_status(c);
	if (cs == STATUS_ACTIVE) {
	    s = STATUS_ACTIVE;
	}
	if (cs == STATUS_BROKEN) {
	    s = STATUS_BROKEN;
	    break;
	}
    }
    cg->status = s;
    return FALSE;
}

extern Status
cgroup_get_status(CGroup g) {
    return g->status;
}

extern void
cgroup_check_status(Manager *manager) {
    g_tree_foreach(manager->cgroups, check_status_fe, manager);
}

static gboolean
cgroup_list_fe(gpointer key,
	gpointer value,
	gpointer data) {
    CGroup cg = value;
    GList **l = data;

    *l = g_list_prepend(*l, g_strdup(cg->name));
    return FALSE;
}

extern GList *
cgroup_list(Manager *manager) {
    GList *l = NULL;

    g_tree_foreach(manager->cgroups, cgroup_list_fe, &l);
    config_file_write(manager->config, NULL);
    return l;
}

extern gboolean
cgroup_add(Manager *manager,
	const SpecCgroup *spec,
	GError **error) {
    CGroup cg;
    GList *l;

    g_assert(manager);
    g_assert(spec);

    cg = g_tree_lookup(manager->cgroups, spec->name);
    if (cg != NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console group %s already exists"), cg->name);
	return FALSE;
    }

    cg = g_new0(struct _CGroup, 1);
    cg->name = g_strdup(spec->name);
    cg->manager = manager;
    cg->config_path = NULL;
    cg->consoles = NULL;
    for (l = spec->consoles; l != NULL; l = g_list_next(l)) {
	cg->consoles = g_list_prepend(cg->consoles, g_strdup(l->data));
    }
    cg->status = STATUS_OK;
    g_tree_insert(manager->cgroups, g_strdup(cg->name), cg);
    cgroup_config_write(cg, TRUE);

    event_raise(manager, EVENT_CGROUP_ADDED, cg->name);
    return TRUE;
}

extern gboolean
cgroup_edit(Manager *manager,
	const gchar *cgroup,
	const SpecCgroup *spec,
	GError **error) {
    CGroup cg;
    gchar *old_name;
    gpointer on_p, cg_p;
    GList *l, *m;
    gboolean changed = FALSE;
    gboolean name_changed = FALSE;
    gboolean members_changed = FALSE;

    g_assert(manager);
    g_assert(cgroup);
    g_assert(spec);

    if (strcmp(cgroup, "ALL") == 0) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_PERMISSION_DENIED,
		_("console group ALL cannot be changed"));
	return FALSE;
    }
    if (strcmp(cgroup, spec->name)) {
	cg = g_tree_lookup(manager->cgroups, spec->name);
	if (cg != NULL) {
	    g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		    _("console group %s already exists"), cg->name);
	    return FALSE;
	}
	changed = TRUE;
	name_changed = TRUE;
    }
    if (g_tree_lookup_extended(manager->cgroups, cgroup,
		&on_p, &cg_p) == FALSE) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console group %s does not exist"), cgroup);
	return FALSE;
    }
    old_name = on_p;
    cg = cg_p;

    if (name_changed) {
	g_free(cg->name);
	cg->name = g_strdup(spec->name);
	config_unset(cg->config_path, "name");
	config_set(cg->config_path, "name", cg->name);
    }

    for (l = spec->consoles; l != NULL; l = g_list_next(l)) {
	if (group_member(cg, l->data) == NULL) {
	    add_console(cg, l->data);
	    members_changed = TRUE;
	}
    }
    l = cg->consoles;
    while (l != NULL) {
	for (l = cg->consoles; l != NULL; l = g_list_next(l)) {
	    for (m = spec->consoles; m != NULL; m = g_list_next(m)) {
		if (!strcmp(l->data, m->data)) {
		    break;
		}
	    }
	    if (m == NULL) {
		remove_console(cg, l->data);
		members_changed = TRUE;
		break;
	    }
	}
    }

    if (members_changed) {
	cgroup_config_write(cg, FALSE);
    }
    if (changed) {
	config_file_write(manager->config, NULL);
    }
    if (name_changed) {
	g_tree_steal(manager->cgroups, old_name);
	g_free(old_name);
	g_tree_insert(manager->cgroups, g_strdup(cg->name), cg);
	event_raise(manager, EVENT_CGROUP_RENAMED, cgroup, cg->name);
    }
    return TRUE;
}


extern gboolean
cgroup_remove(Manager *manager,
	const gchar *cgroup,
	GError **error) {
    CGroup cg;

    g_assert(manager);
    g_assert(cgroup);

    if (strcmp(cgroup, "ALL") == 0) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_PERMISSION_DENIED,
		_("console group ALL cannot be removed"));
	return FALSE;
    }
    cg = g_tree_lookup(manager->cgroups, cgroup);
    if (cg == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console group %s does not exist"), cgroup);
	return FALSE;
    }
    config_delete(cg->config_path);
    cg->config_path = NULL;
    g_tree_remove(manager->cgroups, cgroup);
    config_file_write(manager->config, NULL);
    event_raise(manager, EVENT_CGROUP_REMOVED, cgroup);
    return TRUE;
}

extern SpecCgroup *
cgroup_spec(Manager *manager,
	const gchar *cgroup,
	GError **error) {
    CGroup cg;
    SpecCgroup *sg;
    GList *l;

    g_assert(manager);
    g_assert(cgroup);

    cg = g_tree_lookup(manager->cgroups, cgroup);
    if (cg == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console group %s does not exist"), cgroup);
	return FALSE;
    }
    sg = g_new0(struct _SpecCgroup, 1);
    sg->name = g_strdup(cg->name);
    sg->consoles = NULL;
    for (l = cg->consoles ; l != NULL; l = g_list_next(l)) {
	sg->consoles = g_list_prepend(sg->consoles, g_strdup(l->data));
    }
    return sg;
}

==> conc-2.0.3.tar.gz: conc-2.0.3/mgr/cgroup.c <==
#include <string.h>
#include "cgroup.h"
#include "console.h"
#include "config-file.h"

struct _CGroup {
    gchar *name;
    Manager *manager;
    ConfigPath config_path;
    GList *consoles;
    Status status;
};

static void
cgroup_hash_free(gpointer data) {
    CGroup cg = data;

    if (cg->name != NULL) {
	g_free(cg->name);
    }
    if (cg->config_path != NULL) {
	config_path_free(cg->config_path);
    }
    while (cg->consoles != NULL) {
	g_free(cg->consoles->data);
	cg->consoles = g_list_remove(cg->consoles, cg->consoles->data);
    }
}

static gchar *
group_member(CGroup cg,
	const gchar *console) {
    GList *l;

    for (l = cg->consoles; l; l = g_list_next(l)) {
	if (!strcmp(l->data, console)) {
	    return l->data;
	}
    }
    return NULL;
}

static void
cgroup_config_write(CGroup cg, gboolean flush) {
    ConfigPath cgroups;
    GList *l;

    g_assert(cg);

    if (strcmp(cg->name, "ALL") == 0) {
	return;
    }
    cgroups = config_path(cg->manager->config, "cgroups", NULL);
    if (cgroups == NULL) {
	ConfigPath root = config_path(cg->manager->config, NULL);

	cgroups = config_path_create(cg->manager->config, root, "cgroups");
	config_path_free(root);
    }
    if (cg->config_path != NULL) {
	config_delete(cg->config_path);
    }
    cg->config_path = config_path_create(cg->manager->config,
	    cgroups, "cgroup");
    config_path_free(cgroups);
    config_set(cg->config_path, "name", cg->name);
    for (l = cg->consoles; l != NULL; l = g_list_next(l)) {
	config_set(cg->config_path, "console", l->data);
    }
    if (flush) {
	config_file_write(cg->manager->config, NULL);
    }
}

static void
add_console(CGroup cg,
	const gchar *console) {
    g_assert(cg);
    g_assert(console);

    if (group_member(cg, console) != NULL) {
	return;
    }
    cg->consoles = g_list_prepend(cg->consoles, g_strdup(console));
    event_raise(cg->manager, EVENT_CGROUP_CHANGED, cg->name, console, 1);
}

static void
remove_console(CGroup cg,
	const gchar *console) {
    gchar *m;
    gchar *c = g_strdup(console);

    g_assert(cg);
    g_assert(console);

    m = group_member(cg, console);
    if (m == NULL) {
	return;
    }
    cg->consoles = g_list_remove(cg->consoles, m);
    g_free(m);
    event_raise(cg->manager, EVENT_CGROUP_CHANGED, cg->name, c, 0);
    g_free(c);
}

static void
console_rename(CGroup cg,
	const gchar *oldname,
	const gchar *newname) {
    gchar *m;

    g_assert(cg);
    g_assert(oldname);
    g_assert(newname);

    m = group_member(cg, oldname);
    if (m) {
	cg->consoles = g_list_remove(cg->consoles, m);
	g_free(m);
	cg->consoles = g_list_prepend(cg->consoles, g_strdup(newname));
	cgroup_config_write(cg, TRUE);
    }
}

static void
console_added(Manager *manager,
	const EventData *event,
	gpointer data) {
    CGroup all;

    all = g_hash_table_lookup(manager->cgroups, "ALL");
    g_assert(all);
    add_console(all, event->console_added.console);
    cgroup_config_write(all, TRUE);
}

static void
console_removed_fe(gpointer key,
	gpointer value,
	gpointer data) {
    CGroup cg = value;
    gchar *console = data;

    remove_console(cg, console);
    cgroup_config_write(cg, FALSE);
}

static void
console_removed(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_hash_table_foreach(manager->cgroups, console_removed_fe,
	    (gchar *)event->console_removed.console);
}

static void
console_renamed_fe(gpointer key,
	gpointer value,
	gpointer data) {
    CGroup cg = value;
    EventData *event = data;

    console_rename(cg, event->console_renamed.console,
	    event->console_renamed.newname);
}

static void
console_renamed(Manager *manager,
	const EventData *event,
	gpointer data) {
    g_hash_table_foreach(manager->cgroups, console_renamed_fe,
	    (gpointer)event);
}

extern void
cgroup_load(Manager *manager) {
    ConfigPath cp;
    GList *cgroups, *l;
    CGroup all;


    g_assert(manager);
    g_assert(manager->config);

    manager->cgroups = g_hash_table_new_full(g_str_hash, g_str_equal,
	    g_free, cgroup_hash_free);

    cp = config_path(manager->config, "cgroups", NULL);
    if (cp == NULL) {
	cgroups = NULL;
    } else {
	cgroups = config_children(cp);
	config_path_free(cp);
    }
    for (l = cgroups; l != NULL; l = g_list_next(l)) {
	const gchar *str;
	GList *consoles;
	CGroup cg;

	cp = l->data;
	str = config_get_value(cp, "name");
	if (str == NULL) {
	    g_printerr("configuration file corrupt "
		    "(anonymous console group)\n");
	    continue;
	}
	if (strcmp(str, "ALL") == 0) {
	    /* ALL is reconstructed later */
	    continue;
	}
	cg = g_new0(struct _CGroup, 1);
	cg->name = g_strdup(str);
	cg->manager = manager;
	cg->config_path = cp;
	cg->consoles = NULL;
	cg->status = STATUS_OK;

	consoles = config_children(cp);
	while (consoles != NULL) {
	    ConfigPath ccp = consoles->data;
	    const ConfigDirective *cd = config_get(ccp);

	    consoles = g_list_remove(consoles, ccp);
	    config_path_free(ccp);
	    if (strcmp(cd->name, "console") != 0 ||
		    cd->value == NULL) {
		continue;
	    }
	    cg->consoles = g_list_prepend(cg->consoles, g_strdup(cd->value));
	}
	g_hash_table_insert(manager->cgroups, g_strdup(cg->name), cg);
    }

    all = g_new0(struct _CGroup, 1);
    all->name = g_strdup("ALL");
    all->manager = manager;
    all->config_path = NULL;
    all->consoles = console_list(manager);
    g_hash_table_insert(manager->cgroups, g_strdup("ALL"), all);

    event_register(manager, EVENT_CONSOLE_ADDED, console_added, NULL);
    event_register(manager, EVENT_CONSOLE_REMOVED, console_removed, NULL);
    event_register(manager, EVENT_CONSOLE_RENAMED, console_renamed, NULL);
}

extern GList *
cgroup_list_consoles(CGroup g) {
    GList *p = NULL;
    GList *l;

    for (l = g->consoles; l != NULL; l = g_list_next(l)) {
	p = g_list_prepend(p, g_strdup(l->data));
    }
    return p;
}

static void
check_status_fe(gpointer key,
        gpointer value,
        gpointer data) {
    Manager *manager = data;
    CGroup cg = value;
    GList *l;
    Status s = STATUS_OK;

    for (l = cg->consoles; l != NULL; l = g_list_next(l)) {
	Status cs;
	Console c = g_hash_table_lookup(manager->consoles, l->data);

	if (c == NULL) {
	    continue;
	}
	cs = console_get_status(c);
	if (cs == STATUS_ACTIVE) {
	    s = STATUS_ACTIVE;
	}
	if (cs == STATUS_BROKEN) {
	    s = STATUS_BROKEN;
	    break;
	}
    }
    cg->status = s;
}

extern Status
cgroup_get_status(CGroup g) {
    return g->status;
}

extern void
cgroup_check_status(Manager *manager) {
    g_hash_table_foreach(manager->cgroups, check_status_fe, manager);
}

static void
cgroup_list_fe(gpointer key,
	gpointer value,
	gpointer data) {
    CGroup cg = value;
    GList **l = data;

    *l = g_list_prepend(*l, g_strdup(cg->name));
}

extern GList *
cgroup_list(Manager *manager) {
    GList *l = NULL;

    g_hash_table_foreach(manager->cgroups, cgroup_list_fe, &l);
    config_file_write(manager->config, NULL);
    return l;
}

extern gboolean
cgroup_add(Manager *manager,
	const SpecCgroup *spec,
	GError **error) {
    CGroup cg;
    GList *l;

    g_assert(manager);
    g_assert(spec);

    cg = g_hash_table_lookup(manager->cgroups, spec->name);
    if (cg != NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console group %s already exists"), cg->name);
	return FALSE;
    }

    cg = g_new0(struct _CGroup, 1);
    cg->name = g_strdup(spec->name);
    cg->manager = manager;
    cg->config_path = NULL;
    cg->consoles = NULL;
    for (l = spec->consoles; l != NULL; l = g_list_next(l)) {
	cg->consoles = g_list_prepend(cg->consoles, g_strdup(l->data));
    }
    cg->status = STATUS_OK;
    g_hash_table_insert(manager->cgroups, g_strdup(cg->name), cg);
    cgroup_config_write(cg, TRUE);

    event_raise(manager, EVENT_CGROUP_ADDED, cg->name);
    return TRUE;
}

extern gboolean
cgroup_edit(Manager *manager,
	const gchar *cgroup,
	const SpecCgroup *spec,
	GError **error) {
    CGroup cg;
    gchar *old_name;
    gpointer on_p, cg_p;
    GList *l, *m;
    gboolean changed = FALSE;
    gboolean name_changed = FALSE;
    gboolean members_changed = FALSE;

    g_assert(manager);
    g_assert(cgroup);
    g_assert(spec);

    if (strcmp(cgroup, "ALL") == 0) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_PERMISSION_DENIED,
		_("console group ALL cannot be changed"));
	return FALSE;
    }
    if (strcmp(cgroup, spec->name)) {
	cg = g_hash_table_lookup(manager->cgroups, spec->name);
	if (cg != NULL) {
	    g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		    _("console group %s already exists"), cg->name);
	    return FALSE;
	}
	changed = TRUE;
	name_changed = TRUE;
    }
    if (g_hash_table_lookup_extended(manager->cgroups, cgroup,
		&on_p, &cg_p) == FALSE) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console group %s does not exist"), cgroup);
	return FALSE;
    }
    old_name = on_p;
    cg = cg_p;

    if (name_changed) {
	g_free(cg->name);
	cg->name = g_strdup(spec->name);
	config_unset(cg->config_path, "name");
	config_set(cg->config_path, "name", cg->name);
    }

    for (l = spec->consoles; l != NULL; l = g_list_next(l)) {
	if (group_member(cg, l->data) == NULL) {
	    add_console(cg, l->data);
	    members_changed = TRUE;
	}
    }
    l = cg->consoles;
    while (l != NULL) {
	for (l = cg->consoles; l != NULL; l = g_list_next(l)) {
	    for (m = spec->consoles; m != NULL; m = g_list_next(m)) {
		if (!strcmp(l->data, m->data)) {
		    break;
		}
	    }
	    if (m == NULL) {
		remove_console(cg, l->data);
		members_changed = TRUE;
		break;
	    }
	}
    }

    if (members_changed) {
	cgroup_config_write(cg, FALSE);
    }
    if (changed) {
	config_file_write(manager->config, NULL);
    }
    if (name_changed) {
	g_hash_table_steal(manager->cgroups, old_name);
	g_free(old_name);
	g_hash_table_insert(manager->cgroups, g_strdup(cg->name), cg);
	event_raise(manager, EVENT_CGROUP_RENAMED, cgroup, cg->name);
    }
    return TRUE;
}


extern gboolean
cgroup_remove(Manager *manager,
	const gchar *cgroup,
	GError **error) {
    CGroup cg;

    g_assert(manager);
    g_assert(cgroup);

    if (strcmp(cgroup, "ALL") == 0) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_PERMISSION_DENIED,
		_("console group ALL cannot be removed"));
	return FALSE;
    }
    cg = g_hash_table_lookup(manager->cgroups, cgroup);
    if (cg == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console group %s does not exist"), cgroup);
	return FALSE;
    }
    config_delete(cg->config_path);
    cg->config_path = NULL;
    g_hash_table_remove(manager->cgroups, cgroup);
    config_file_write(manager->config, NULL);
    event_raise(manager, EVENT_CGROUP_REMOVED, cgroup);
    return TRUE;
}

extern SpecCgroup *
cgroup_spec(Manager *manager,
	const gchar *cgroup,
	GError **error) {
    CGroup cg;
    SpecCgroup *sg;
    GList *l;

    g_assert(manager);
    g_assert(cgroup);

    cg = g_hash_table_lookup(manager->cgroups, cgroup);
    if (cg == NULL) {
	g_set_error(error, MANAGER_ERROR, MANAGER_ERROR_DOES_NOT_EXIST,
		_("console group %s does not exist"), cgroup);
	return FALSE;
    }
    sg = g_new0(struct _SpecCgroup, 1);
    sg->name = g_strdup(cg->name);
    sg->consoles = NULL;
    for (l = cg->consoles ; l != NULL; l = g_list_next(l)) {
	sg->consoles = g_list_prepend(sg->consoles, g_strdup(l->data));
    }
    return sg;
}
